@@ -38,10 +38,12 @@
  *
  */
 
+#define _GNU_SOURCE
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
 
+#define _GNU_SOURCE
 #include <stdio.h>
 #include <unistd.h>
 #include <string.h>
@@ -52,6 +54,9 @@
 #include <stdarg.h>
 #include <assert.h>
 #include <sys/time.h>
+#include <sys/file.h>
+#include <sys/socket.h>
+#include <sys/un.h>
 
 #include <pulse/pulseaudio.h>
 #include <pulse/error.h>
@@ -61,7 +66,6 @@
 
 #include "pacat-simple-vchan.h"
 #include "qubes-vchan-sink.h"
-#include "pacat-control-object.h"
 
 #define CLEAR_LINE "\x1B[K"
 #ifdef __GNUC__
@@ -580,6 +584,223 @@ fail:
 
 }
 
+/*
+ * domid: remote end domid (IN)
+ * pidfile_path: path to the created file (OUT)
+ * pidfile_fd: open FD of the pidfile, used also as a lockfile (OUT)
+ *
+ * Returns: 0 on success -1 otherwise
+ */
+static int create_pidfile(int domid, char **pidfile_path, int *pidfile_fd)
+{
+    int fd, ret;
+    char pid_s[16];
+
+    if (asprintf(pidfile_path, PACAT_PIDFILE_PATH_TPL, domid) < 0) {
+        pacat_log("Failed to construct pidfile path, out of memory?");
+        return -1;
+    }
+
+    fd = open(*pidfile_path, O_RDWR | O_CREAT | O_CLOEXEC, 0644);
+    if (fd < 0) {
+        pacat_log("Failed to create pidfile %s: %s", *pidfile_path, strerror(errno));
+        return -1;
+    }
+
+    if (flock(fd, LOCK_EX | LOCK_NB) < 0) {
+        if (errno == EWOULDBLOCK) {
+            ret = read(fd, pid_s, sizeof(pid_s)-1);
+            if (ret <= 0) {
+                pacat_log("Another instance of pacat-simple-vchan is already running for this VM, but failed to get its PID");
+            } else {
+                pid_s[ret] = '\0';
+                if (pid_s[ret-1] == '\n')
+                    pid_s[ret-1] = '\0';
+                pacat_log("Another instance of pacat-simple-vchan is already running for this VM (PID: %s)", pid_s);
+            }
+        } else {
+            pacat_log("Failed to obtain a lock on the pid file: %s", pa_strerror(errno));
+        }
+        close(fd);
+        return -1;
+    }
+
+    ret = snprintf(pid_s, sizeof(pid_s), "%d\n", getpid());
+    if (write(fd, pid_s, ret) != ret) {
+        pacat_log("Failed to write a pid file: %s", pa_strerror(errno));
+        close(fd);
+        return -1;
+    }
+
+    *pidfile_fd = fd;
+    return 0;
+}
+
+static void check_vchan_eof_timer(pa_mainloop_api*a, pa_time_event* e,
+               const struct timeval *UNUSED(tv), void *userdata)
+{
+    struct userdata *u = userdata;
+    struct timeval restart_tv;
+    assert(u);
+
+    if (!libvchan_is_open(u->play_ctrl)) {
+        pacat_log("vchan_is_eof (timer)");
+        quit(u, 0);
+        return;
+    }
+
+    pa_gettimeofday(&restart_tv);
+    pa_timeval_add(&restart_tv, (pa_usec_t) 5 * 1000 * PA_USEC_PER_MSEC);
+    a->time_restart(e, &restart_tv);
+}
+
+static void control_socket_callback(pa_mainloop_api *UNUSED(a),
+        pa_io_event *UNUSED(e), int fd, pa_io_event_flags_t f,
+        void *userdata) {
+    struct userdata *u = userdata;
+    int client_fd;
+    char command_buffer[32];
+    size_t command_len = 0;
+    int ret;
+    int new_rec_allowed = -1;
+
+    if (!(f & PA_IO_EVENT_INPUT))
+        return;
+
+    client_fd = accept(fd, NULL, NULL);
+    if (client_fd < 0) {
+        pacat_log("Accept control connection failed: %s", strerror(errno));
+        return;
+    }
+
+    /* read until either:
+     *  - end of command (\n) is found
+     *  - EOF
+     */
+    do {
+        ret = read(client_fd, command_buffer+command_len, sizeof(command_buffer)-command_len);
+        if (ret < 0) {
+            pacat_log("Control client read failed: %s", strerror(errno));
+        }
+        command_len += ret;
+        if (ret == 0)
+            break;
+    } while (!memchr(command_buffer, '\n', command_len));
+
+    if (strncmp(command_buffer, "audio-input 0\n", command_len) == 0) {
+        new_rec_allowed = 0;
+    } else if (strncmp(command_buffer, "audio-input 1\n", command_len) == 0) {
+        new_rec_allowed = 1;
+    }
+    if (new_rec_allowed != -1) {
+        g_mutex_lock(&u->prop_mutex);
+        u->rec_allowed = new_rec_allowed;
+        pacat_log("Setting audio-input to %s", u->rec_allowed ? "enabled" : "disabled");
+        if (u->rec_allowed && u->rec_requested) {
+            pacat_log("Recording start");
+            pa_stream_cork(u->rec_stream, 0, NULL, NULL);
+        } else if (!u->rec_allowed &&
+                (u->rec_requested || !pa_stream_is_corked(u->rec_stream))) {
+            pacat_log("Recording stop");
+            pa_stream_cork(u->rec_stream, 1, NULL, NULL);
+        }
+        g_mutex_unlock(&u->prop_mutex);
+        if (!qdb_write(u->qdb, u->qdb_path, new_rec_allowed ? "1" : "0", 1)) {
+            pacat_log("Failed to write QubesDB %s: %s", u->qdb_path, strerror(errno));
+        }
+    }
+    /* accept only one command per connection */
+    close(client_fd);
+}
+
+static int setup_control(struct userdata *u) {
+    int socket_fd = -1;
+    struct sockaddr_un addr;
+
+    socket_fd = socket(AF_UNIX, SOCK_STREAM, 0);
+    if (socket_fd == -1) {
+        pacat_log("socket failed: %s", strerror(errno));
+        goto fail;
+    }
+
+    if (snprintf(addr.sun_path, sizeof(addr.sun_path),
+                "/var/run/qubes/audio-control.%s", u->name)
+            >= (int)sizeof(addr.sun_path)) {
+        pacat_log("VM name too long");
+        goto fail;
+    }
+
+    /* ignore result */
+    unlink(addr.sun_path);
+
+    if (bind(socket_fd, &addr, sizeof(addr)) == -1) {
+        pacat_log("bind to %s failed: %s", addr.sun_path, strerror(errno));
+        goto fail;
+    }
+
+    if (listen(socket_fd, 5) == -1) {
+        pacat_log("listen on %s failed: %s", addr.sun_path, strerror(errno));
+        goto fail;
+    }
+
+    u->control_socket_event = u->mainloop_api->io_new(u->mainloop_api,
+            socket_fd, PA_IO_EVENT_INPUT, control_socket_callback, u);
+    if (!u->control_socket_event) {
+        pacat_log("io_new control failed");
+        goto fail;
+    }
+
+    u->qdb = qdb_open(NULL);
+    if (!u->qdb) {
+        pacat_log("qdb_open failed: %s", strerror(errno));
+        goto fail;
+    }
+
+    if (asprintf(&u->qdb_path, "/audio-input/%s", u->name) < 0) {
+        pacat_log("QubesDB path setup failed: %s", strerror(errno));
+        u->qdb_path = NULL;
+        goto fail;
+    }
+
+    if (!qdb_write(u->qdb, u->qdb_path, "0", 1)) {
+        pacat_log("qdb_write failed: %s", strerror(errno));
+        goto fail;
+    }
+
+    u->control_socket_fd = socket_fd;
+
+    return 0;
+
+fail:
+    if (u->qdb_path)
+        free(u->qdb_path);
+    u->qdb_path = NULL;
+    if (u->qdb)
+        qdb_close(u->qdb);
+    u->qdb = NULL;
+    if (u->control_socket_event)
+        u->mainloop_api->io_free(u->control_socket_event);
+    u->control_socket_event = NULL;
+    if (socket_fd >= 0)
+        close(socket_fd);
+
+    return 1;
+}
+
+static void control_cleanup(struct userdata *u) {
+
+    if (u->control_socket_event)
+        u->mainloop_api->io_free(u->control_socket_event);
+    if (u->control_socket_fd > 0)
+        close(u->control_socket_fd);
+    if (u->qdb && u->qdb_path)
+        qdb_rm(u->qdb, u->qdb_path);
+    if (u->qdb_path)
+        free(u->qdb_path);
+    if (u->qdb)
+        qdb_close(u->qdb);
+}
+
 int main(int argc, char *argv[])
 {
     struct timeval tv;
@@ -589,6 +810,8 @@ int main(int argc, char *argv[])
     char *server = NULL;
     char *domname = NULL;
     int domid = -1;
+    char *pidfile_path;
+    int pidfile_fd;
     int i;
 
     if (argc <= 2) {
@@ -622,6 +845,10 @@ int main(int argc, char *argv[])
         exit(1);
     }
 
+    if (create_pidfile(domid, &pidfile_path, &pidfile_fd) < 0)
+        /* error already printed by create_pidfile() */
+        exit(1);
+
     memset(&u, 0, sizeof(u));
     u.ret = 1;
 
@@ -680,6 +907,12 @@ int main(int argc, char *argv[])
 
     pa_gettimeofday(&tv);
     pa_timeval_add(&tv, (pa_usec_t) 5 * 1000 * PA_USEC_PER_MSEC);
+    time_event = u.mainloop_api->time_new(u.mainloop_api, &tv, check_vchan_eof_timer, &u);
+    if (!time_event) {
+        pacat_log("time_event create failed");
+        goto quit;
+    }
+
     if (!(u.context = pa_context_new_with_proplist(u.mainloop_api, NULL, u.proplist))) {
         pacat_log("pa_context_new() failed.");
         goto quit;
@@ -692,8 +925,8 @@ int main(int argc, char *argv[])
         goto quit;
     }
 
-    if (dbus_init(&u) < 0) {
-        pacat_log("dbus initialization failed");
+    if (setup_control(&u) < 0) {
+        pacat_log("control socket initialization failed");
         goto quit;
     }
 
@@ -703,15 +936,10 @@ int main(int argc, char *argv[])
     g_main_loop_run (u.loop);
 
 quit:
-    if (u.pacat_control) {
-        assert(u.dbus);
-        dbus_g_connection_unregister_g_object(u.dbus, u.pacat_control);
-        g_object_unref(u.pacat_control);
+    if (u.control_socket_event) {
+        control_cleanup(&u);
     }
 
-    if (u.dbus)
-        dbus_g_connection_unref(u.dbus);
-
     if (u.play_stream)
         pa_stream_unref(u.play_stream);
 
@@ -762,5 +990,8 @@ quit:
         pa_proplist_free(u.proplist);
 
     g_mutex_clear(&u.prop_mutex);
+
+    unlink(pidfile_path);
+    close(pidfile_fd);
     return u.ret;
 }
